@@ -67,6 +67,11 @@ class StringTypeTests(TypeTestHelper, unittest.TestCase):
         self.type_instance = types.String(choices=('foo', 'bar'))
         self.assertConvertedValue('foo', 'foo')
 
+    def test_listed_value_dict(self):
+        self.type_instance = types.String(choices=[
+            ('foo', 'ab'), ('bar', 'xy')])
+        self.assertConvertedValue('foo', 'foo')
+
     def test_unlisted_value(self):
         self.type_instance = types.String(choices=['foo', 'bar'])
         self.assertInvalid('baz')
@@ -88,6 +93,11 @@ class StringTypeTests(TypeTestHelper, unittest.TestCase):
         self.type_instance = types.String(quotes=True)
         self.assertConvertedValue('foo bar"', 'foo bar"')
 
+    def test_single_quote_is_invalid(self):
+        self.type_instance = types.String(quotes=True)
+        self.assertInvalid('"')
+        self.assertInvalid("'")
+
     def test_repr(self):
         t = types.String()
         self.assertEqual('String', repr(t))
@@ -98,7 +108,11 @@ class StringTypeTests(TypeTestHelper, unittest.TestCase):
 
     def test_repr_with_choices_tuple(self):
         t = types.String(choices=('foo', 'bar'))
-        self.assertEqual('String(choices=(\'foo\', \'bar\'))', repr(t))
+        self.assertEqual('String(choices=[\'foo\', \'bar\'])', repr(t))
+
+    def test_repr_with_choices_dict(self):
+        t = types.String(choices=[('foo', 'ab'), ('bar', 'xy')])
+        self.assertEqual('String(choices=[\'foo\', \'bar\'])', repr(t))
 
     def test_equal(self):
         self.assertTrue(types.String() == types.String())
@@ -108,9 +122,8 @@ class StringTypeTests(TypeTestHelper, unittest.TestCase):
         t2 = types.String(choices=['foo', 'bar'])
         t3 = types.String(choices=('foo', 'bar'))
         t4 = types.String(choices=['bar', 'foo'])
-        self.assertTrue(t1 == t2)
-        self.assertTrue(t1 == t3)
-        self.assertTrue(t1 == t4)
+        t5 = types.String(choices=[('foo', 'ab'), ('bar', 'xy')])
+        self.assertTrue(t1 == t2 == t3 == t4 == t5)
 
     def test_not_equal_with_different_choices(self):
         t1 = types.String(choices=['foo', 'bar'])
@@ -282,7 +295,11 @@ class IntegerTypeTests(TypeTestHelper, unittest.TestCase):
 
     def test_repr_with_choices_tuple(self):
         t = types.Integer(choices=(80, 457))
-        self.assertEqual('Integer(choices=(80, 457))', repr(t))
+        self.assertEqual('Integer(choices=[80, 457])', repr(t))
+
+    def test_repr_with_choices_dict(self):
+        t = types.Integer(choices=[(80, 'ab'), (457, 'xy')])
+        self.assertEqual('Integer(choices=[80, 457])', repr(t))
 
     def test_equal(self):
         self.assertTrue(types.Integer() == types.Integer())
@@ -302,8 +319,8 @@ class IntegerTypeTests(TypeTestHelper, unittest.TestCase):
         t1 = types.Integer(choices=[80, 457])
         t2 = types.Integer(choices=[457, 80])
         t3 = types.Integer(choices=(457, 80))
-        self.assertTrue(t1 == t2)
-        self.assertTrue(t1 == t3)
+        t4 = types.Integer(choices=[(80, 'ab'), (457, 'xy')])
+        self.assertTrue(t1 == t2 == t3 == t4)
 
     def test_not_equal(self):
         self.assertFalse(types.Integer(min=123) == types.Integer(min=456))
@@ -369,21 +386,24 @@ class IntegerTypeTests(TypeTestHelper, unittest.TestCase):
         self.assertRaises(ValueError, t, 201)
         self.assertRaises(ValueError, t, -457)
 
-    def test_with_choices_list(self):
-        t = types.Integer(choices=[80, 457])
+    def _test_with_choices(self, t):
         self.assertRaises(ValueError, t, 1)
         self.assertRaises(ValueError, t, 200)
         self.assertRaises(ValueError, t, -457)
         t(80)
         t(457)
 
+    def test_with_choices_list(self):
+        t = types.Integer(choices=[80, 457])
+        self._test_with_choices(t)
+
     def test_with_choices_tuple(self):
         t = types.Integer(choices=(80, 457))
-        self.assertRaises(ValueError, t, 1)
-        self.assertRaises(ValueError, t, 200)
-        self.assertRaises(ValueError, t, -457)
-        t(80)
-        t(457)
+        self._test_with_choices(t)
+
+    def test_with_choices_dict(self):
+        t = types.Integer(choices=[(80, 'ab'), (457, 'xy')])
+        self._test_with_choices(t)
 
 
 class FloatTypeTests(TypeTestHelper, unittest.TestCase):
@@ -519,6 +539,10 @@ class ListTypeTests(TypeTestHelper, unittest.TestCase):
                                    'bar, baz',
                                    'bam'])
 
+    def test_list_of_values_containing_trailing_comma(self):
+        self.assertConvertedValue('foo, bar, baz, ',
+                                  ['foo', 'bar', 'baz'])
+
     def test_list_of_lists(self):
         self.type_instance = types.List(
             types.List(types.String(), bounds=True)
@@ -531,6 +555,16 @@ class ListTypeTests(TypeTestHelper, unittest.TestCase):
         self.assertConvertedValue('1,2,3,5',
                                   [1, 2, 3, 5])
 
+    def test_list_of_custom_type_containing_trailing_comma(self):
+        self.type_instance = types.List(types.Integer())
+        self.assertConvertedValue('1,2,3,5,',
+                                  [1, 2, 3, 5])
+
+    def test_tuple_of_custom_type(self):
+        self.type_instance = types.List(types.Integer())
+        self.assertConvertedValue(('1', '2', '3', '5'),
+                                  [1, 2, 3, 5])
+
     def test_bounds_parsing(self):
         self.type_instance = types.List(types.Integer(), bounds=True)
         self.assertConvertedValue('[1,2,3]', [1, 2, 3])
@@ -707,6 +741,32 @@ class IPv6AddressTypeTests(IPAddressTypeTests):
         self.assertInvalid('192.168.0.1')
 
 
+class HostAddressTypeTests(TypeTestHelper, unittest.TestCase):
+    type = types.HostAddress()
+
+    def test_invalid_host_addresses(self):
+        self.assertInvalid('-1')
+        self.assertInvalid('_foo')
+        self.assertInvalid('3.14')
+        self.assertInvalid('10.0')
+        self.assertInvalid('host..name')
+        self.assertInvalid('org.10')
+        self.assertInvalid('0.0.00')
+
+    def test_valid_host_addresses(self):
+        self.assertConvertedValue('foo.bar', 'foo.bar')
+        self.assertConvertedValue('192.168.0.1', '192.168.0.1')
+        self.assertConvertedValue('abcd:ef::1', 'abcd:ef::1')
+        self.assertConvertedValue('home-site-here.org.com',
+                                  'home-site-here.org.com')
+        self.assertConvertedValue('3com.com', '3com.com')
+        self.assertConvertedValue('10.org', '10.org')
+        self.assertConvertedValue('cell1.nova.site1', 'cell1.nova.site1')
+        self.assertConvertedValue('ab-c.com', 'ab-c.com')
+        self.assertConvertedValue('abc.com-org', 'abc.com-org')
+        self.assertConvertedValue('abc.0-0', 'abc.0-0')
+
+
 class HostnameTypeTests(TypeTestHelper, unittest.TestCase):
     type = types.Hostname()
 
@@ -746,6 +806,12 @@ class HostnameTypeTests(TypeTestHelper, unittest.TestCase):
         self.assertInvalid(".host.name.com")
         self.assertInvalid("no spaces")
 
+    def test_invalid_hostnames_with_numeric_characters(self):
+        self.assertInvalid("10.0.0.0")
+        self.assertInvalid("3.14")
+        self.assertInvalid("org.10")
+        self.assertInvalid('0.0.00')
+
     def test_no_start_end_hyphens(self):
         self.assertInvalid("-host.com")
         self.assertInvalid("-hostname.com-")
@@ -759,9 +825,13 @@ class HostnameTypeTests(TypeTestHelper, unittest.TestCase):
         self.assertConvertedEqual('cell1.nova.site1')
         self.assertConvertedEqual('site01001')
         self.assertConvertedEqual('home-site-here.org.com')
-        self.assertConvertedEqual('192.168.0.1')
-        self.assertConvertedEqual('1.1.1')
         self.assertConvertedEqual('localhost')
+        self.assertConvertedEqual('3com.com')
+        self.assertConvertedEqual('10.org')
+        self.assertConvertedEqual('10ab.10ab')
+        self.assertConvertedEqual('ab-c.com')
+        self.assertConvertedEqual('abc.com-org')
+        self.assertConvertedEqual('abc.0-0')
 
     def test_max_segment_size(self):
         self.assertConvertedEqual('host.%s.com' % ('x' * 63))
@@ -791,6 +861,35 @@ class URITypeTests(TypeTestHelper, unittest.TestCase):
         self.assertConvertedValue('http://www.example.com',
                                   'http://www.example.com')
 
+    def test_equality(self):
+        a = types.URI()
+        b = types.URI()
+        self.assertEqual(a, b)
+
+    def test_equality_length(self):
+        a = types.URI(max_length=5)
+        b = types.URI(max_length=5)
+        self.assertEqual(a, b)
+
+    def test_equality_length_not(self):
+        a = types.URI()
+        b = types.URI(max_length=5)
+        c = types.URI(max_length=10)
+        self.assertNotEqual(a, b)
+        self.assertNotEqual(c, b)
+
+    def test_equality_schemes(self):
+        a = types.URI(schemes=['ftp'])
+        b = types.URI(schemes=['ftp'])
+        self.assertEqual(a, b)
+
+    def test_equality_schemes_not(self):
+        a = types.URI()
+        b = types.URI(schemes=['ftp'])
+        c = types.URI(schemes=['http'])
+        self.assertNotEqual(a, b)
+        self.assertNotEqual(c, b)
+
 
 class PortTypeTests(TypeTestHelper, unittest.TestCase):
     type = types.Port()
@@ -824,16 +923,29 @@ class PortTypeTests(TypeTestHelper, unittest.TestCase):
 
     def test_repr_with_choices_tuple(self):
         t = types.Port(choices=(80, 457))
-        self.assertEqual('Port(choices=(80, 457))', repr(t))
+        self.assertEqual('Port(choices=[80, 457])', repr(t))
 
-    def test_choices(self):
-        t = types.Port(choices=[80, 457])
+    def _test_with_choices(self, t):
         self.assertRaises(ValueError, t, 1)
         self.assertRaises(ValueError, t, 200)
+        self.assertRaises(ValueError, t, -457)
         t(80)
         t(457)
 
+    def test_with_choices_list(self):
+        t = types.Port(choices=[80, 457])
+        self._test_with_choices(t)
+
+    def test_with_choices_tuple(self):
+        t = types.Port(choices=(80, 457))
+        self._test_with_choices(t)
+
+    def test_with_choices_dict(self):
+        t = types.Port(choices=[(80, 'ab'), (457, 'xy')])
+        self._test_with_choices(t)
+
     def test_invalid_choices(self):
+        """Check for choices that are specifically invalid for ports."""
         self.assertRaises(ValueError, types.Port, choices=[-1, 457])
         self.assertRaises(ValueError, types.Port, choices=[1, 2, 3, 65536])
 
@@ -855,8 +967,8 @@ class PortTypeTests(TypeTestHelper, unittest.TestCase):
         t1 = types.Port(choices=[80, 457])
         t2 = types.Port(choices=[457, 80])
         t3 = types.Port(choices=(457, 80))
-        self.assertTrue(t1 == t2)
-        self.assertTrue(t1 == t3)
+        t4 = types.Port(choices=[(457, 'ab'), (80, 'xy')])
+        self.assertTrue(t1 == t2 == t3 == t4)
 
     def test_not_equal(self):
         self.assertFalse(types.Port(min=123) == types.Port(min=456))
@@ -933,18 +1045,39 @@ class PortTypeTests(TypeTestHelper, unittest.TestCase):
         self.assertRaises(ValueError, t, 1)
         t(0)
 
-    def test_with_choices_list(self):
-        t = types.Port(choices=[80, 457])
-        self.assertRaises(ValueError, t, 1)
-        self.assertRaises(ValueError, t, 200)
-        self.assertRaises(ValueError, t, -457)
-        t(80)
-        t(457)
 
-    def test_with_choices_tuple(self):
-        t = types.Port(choices=(80, 457))
-        self.assertRaises(ValueError, t, 1)
-        self.assertRaises(ValueError, t, 200)
-        self.assertRaises(ValueError, t, -457)
-        t(80)
-        t(457)
+class FormatSampleDefaultTests(unittest.TestCase):
+    def test_string(self):
+        t = types.String()
+        self.assertEqual([' bar '],
+                         t.format_defaults('foo', sample_default=' bar '))
+
+    def test_string_non_str(self):
+        t = types.String()
+        e = Exception('bar')
+        self.assertEqual(['bar'],
+                         t.format_defaults('', sample_default=e))
+
+    def test_string_non_str_spaces(self):
+        t = types.String()
+        e = Exception(' bar ')
+        self.assertEqual(['" bar "'],
+                         t.format_defaults('', sample_default=e))
+
+    def test_list_string(self):
+        t = types.List(item_type=types.String())
+        test_list = ['foo', Exception(' bar ')]
+        self.assertEqual(['foo," bar "'],
+                         t.format_defaults('', sample_default=test_list))
+
+    def test_list_no_type(self):
+        t = types.List()
+        test_list = ['foo', Exception(' bar ')]
+        self.assertEqual(['foo," bar "'],
+                         t.format_defaults('', sample_default=test_list))
+
+    def test_list_not_list(self):
+        t = types.List()
+        self.assertEqual(['foo'],
+                         t.format_defaults('',
+                                           sample_default=Exception('foo')))
